{% extends "!layout.html" %}
{% block extrahead %}
<script type="text/javascript" src="https://docs.python.org/3/_static/jquery.js"></script>
<script type="text/javascript" src="https://docs.python.org/3/_static/underscore.js"></script>
<script type="text/javascript" src="https://docs.python.org/3/_static/doctools.js"></script>
<script type="text/javascript" src="https://docs.python.org/3/_static/copybutton.js"></script>
{% endblock %}
