#!/usr/bin/env python
"""Function decorators, exceptions, and several commonly-used that operate
on primitive object types

Package overview
================

    ===============================================  ==============================================
    **Package module**                               **Contents**
    -----------------------------------------------  ----------------------------------------------
    :py:mod:`~plastid.util.services.decorators`         Function decorators
    :py:mod:`~plastid.util.services.exceptions`         Exceptions
    :py:mod:`~plastid.util.services.lists`              Miscellaneous list operations
    :py:mod:`~plastid.util.services.sets`               Miscellaneous set operations
    :py:mod:`~plastid.util.services.misc`               Other miscellaneous functions
    :py:mod:`~plastid.util.services.mini2to3`           Python 2.x/3.x compatability layer
    ===============================================  ==============================================
 
"""
