div.captionfigure p.caption span.caption-text {
	font-size: 14px;
	font-family: "Lato","proxima-nova","Helvetica Neue",Arial,sans-serif;
	font-weight: normal;
	color: #404040;
	line-height: 18px;
}

div.captionfigure p.caption {
	margin-top: 18px;
	margin-bottom: 24px;
}

div.captionfigure image {
	margin-top: 24px;
	margin-bottom: 24px;
}