/* Fixed colors - always in color map. */

/* Copyright (C) 2011 The Regents of the University of California 
 * See kent/LICENSE or http://genome.ucsc.edu/license/ for licensing information. */
#include "common.h"
#include "memgfx.h"


struct rgbColor mgFixedColors[9] = {
/* These correspond to MG_WHITE, MG_BLACK, etc. */
    { 255, 255, 255},
    { 0, 0, 0},
    { 255, 0, 0},
    { 0, 255, 0},
    { 0, 0, 255},
    { 0, 255, 255},
    { 255, 0, 255},
    { 255, 255, 0},
    { 140, 140, 140},
};



