#!/usr/bin/env python
"""Library components for writing command-line scripts

Package overview
================

    =================================================    =========================
    **Package module**                                   **Contents**
    -------------------------------------------------    -------------------------
    :py:mod:`~plastid.util.scriptlib.argparsers`          :class:`~argparse.ArgumentParser` objects for reading and processing various genomics file types 
    :py:mod:`~plastid.util.scriptlib.help_formatters`     Utilities to reformat module docstrings for use as command-line help text
    :py:mod:`~plastid.util.scriptlib.template`            Template for writing a command-line script using :data:`plastid`
    =================================================    =========================    
"""
