/* longTabix -- long range pairwise interaction format, from Wash U
 *    Documented here:  http://wiki.wubrowse.org/Long-range
 */

/* Copyright (C) 2017 The Regents of the University of California 
 * See kent/LICENSE or http://genome.ucsc.edu/license/ for licensing information. */

struct asObject *longTabixAsObj();
// Return asObject describing fields of longTabix file

