/* Copyright (C) 2017 The Regents of the University of California 
 *  * See kent/LICENSE or http://genome.ucsc.edu/license/ for licensing information. */

char *windowsToAscii(char *input);
/* Convert UTF-8 to ascii. */
